#include <boost/math/complex/complex.hpp>
#include <boost/math/constants/constants.hpp>
#include <boost/multiprecision/cpp_dec_float.hpp>
#include "fixed_point.h"

using boost::multiprecision::cpp_dec_float_50;

namespace
{
  typedef enum enum_polynomial_type
  {
    chebyshev_t_polynomial_type = 1,
    chebyshev_u_polynomial_type = 2,
    laguerre_l_polynomial_type  = 3,
    hermite_h_polynomial_type   = 4
  }
  polynomial_type;

  template<typename T>
  T orthogonal_polynomial_template(const T& x,
                                   const unsigned n,
                                   const polynomial_type type)
  {
    // Compute the value of an orthogonal polynomial of one of the following types:
    // Chebyshev 1st, Chebyshev 2nd, Laguerre, or Hermite

    T y0(1);

    if(n == 0U)
    {
      return y0;
    }

    T y1;

    if(type == chebyshev_t_polynomial_type)
    {
      y1 = x;
    }
    else if(type == laguerre_l_polynomial_type)
    {
      y1 = 1 - x;
    }
    else
    {
      y1 =  x * 2;
    }

    if(n == 1U)
    {
      return y1;
    }

    T a(2);
    T b(0);
    T c(1);

    T yk;

    // Calculate higher orders using the recurrence relation.
    // The direction of stability is upward recurrence.
    for(int k = 2; k <= static_cast<int>(n); k++)
    {
      if(type == laguerre_l_polynomial_type)
      {
        a = T::value_type(-1) / k;
        b =  2 + a;
        c =  1 + a;
      }
      else if(type == hermite_h_polynomial_type)
      {
        c = 2 * (k - 1);
      }

      yk = (((a * x) + b) * y1) - (c * y0);

      y0 = y1;
      y1 = yk;
    }

    return yk;
  }

  template<typename T>
  T chebyshev_t(const int n, const T& z)
  {
    if(n < 0)
    {
      return chebyshev_t(-n, z);
    }
    else
    {
      return orthogonal_polynomial_template(z, static_cast<unsigned>(n), chebyshev_t_polynomial_type);
    }
  }

  template<typename T>
  T chebyshev_u(const int n, const T& z)
  {
    if(n < 0)
    {
      if(n == -2)
      {
        return T(1);
      }
      else if(n == -1)
      {
        return T(0);
      }
      else
      {
        return -chebyshev_u(-n - 2, z);
      }
    }
    else
    {
      return orthogonal_polynomial_template(z, static_cast<unsigned>(n), chebyshev_u_polynomial_type);
    }
  }

  template<typename T>
  T laguerre(const int n, const T& z)
  {
    if(z.real() < 0)
    {
      // TBD: Negative argument not yet implemented for complex Laguerre polynomials.
      return T(0);
    }


    if(n < 0)
    {
      // TBD: Negative order not yet implemented for complex Laguerre polynomials.
      return T(0);
    }
    else if(n == 0)
    {
      return T(1);
    }
    else if(n == 1)
    {
      // TBD: boost::math::complex needs global operators with POD.
      return 1 - z;
    }

    return orthogonal_polynomial_template(z, static_cast<unsigned>(n), laguerre_l_polynomial_type);
  }

  template<typename T>
  T hermite(const int n, const T& z)
  {
    if(n < 0)
    {
      // TBD: Negative order not yet implemented for complex Hermite polynomials.
      return T(0);
    }
    else
    {
      return orthogonal_polynomial_template(z, static_cast<unsigned>(n), hermite_h_polynomial_type);
    }
  }
}

namespace
{
  // Define a fixed-point type using my own Q15.16 representation.
  typedef fixed_point_15pt16 fixed_point_type;
}

namespace boost
{
  namespace math
  {
    namespace constants
    {
      // Define a template specialization of boost::math::constants:::pi()
      // for the fixed-point type.
      template<>
      inline fixed_point_type pi<>(void) { return const_pi<fixed_point_type>(); }
    }
  }
}

void do_complex_fixed(void)
{

  const boost::math::complex<fixed_point_type> z(fixed_point_type(2) / 3,
                                                 fixed_point_type(6) / 7);

  boost::math::complex<fixed_point_type> result;

  // 0.9360951 + 0.4578289 I
  result = boost::math::sqrt(z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // 0.8597787 + 0.7591850 I
  result = boost::math::sin(z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // 0.3133628 + 0.8660934 I
  result = boost::math::tan(z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // 1.2749868 + 1.4724390 I
  result = boost::math::exp(z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // 0.082392 + 0.90975 I
  result = boost::math::log(z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // 70.707 + 199.25 I
  result = ::chebyshev_t(7U, z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // 175.084 + 339.563 I
  result = ::chebyshev_u(7U, z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // -1.5997 + 2.28825 I
  result = ::laguerre(7U, z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;

  // -4504.008 + 2305.48 I
  result = ::hermite(7U, z);
  std::cout << std::setprecision(std::numeric_limits<fixed_point_type>::digits10)
            << "(" << result.real().to_float() << "," << result.imag().to_float() << ")"
            << std::endl;
}

void do_complex(void)
{
  const boost::math::complex<cpp_dec_float_50> z(cpp_dec_float_50(2) / 3,
                                                 cpp_dec_float_50(6) / 7);

  // 0.9360950870248013785669224072998760238676132741143621 + 0.4578289476270627861756707972781501074975513632381610 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << boost::math::sqrt(z)
            << std::endl;

  // 0.8597786805880400406516837200086747931320578256393108 + 0.7591850108460004188252800145931283112097383472688717 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << boost::math::sin(z)
            << std::endl;

  // 0.313362812055036781810959306664414769100283931341407 + 0.866093357762006429550957421956067187566578048142006 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << boost::math::tan(z)
            << std::endl;

  // 1.27498683314007904058067490147572815227208614825386 + 1.47243895289502918723938329689559875827478287660417 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << boost::math::exp(z)
            << std::endl;

  // 0.082391968064313522952373589217312539994109563650135 + 0.909753157944209744647015694983746746622462541613581 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << boost::math::log(z)
            << std::endl;

  // 70.7074048582267894180111826051532243910933860080563 + 199.2535168419574104658078550287106623704847611930923 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << ::chebyshev_t(7U, z)
            << std::endl;

  // 175.083874218041566008719612413546525362075467227127 + 339.562590341304048083441779001357823862808086123957 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << ::chebyshev_u(7U, z)
            << std::endl;

  // -1.59965144684750460089186204594875111756713525785022 + 2.28825178485690639268323931412352005440438176183398 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << ::laguerre(7U, z)
            << std::endl;

  // -4504.00773884441697749938657790838723680492285137469 + 2305.48091802223064612790737865229691670110958748252 I
  std::cout << std::setprecision(std::numeric_limits<cpp_dec_float_50>::digits10)
            << ::hermite(7U, z)
            << std::endl;
}
