// complex.hpp header file ------------------------------------- //

//          Copyright Matthieu Schaller 2012
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

// See ../../libs/doc/html/index.html for documentation

// Revision history
// 28 Feb 2012 Initial version (Matthieu Schaller)

#ifndef COMPLEX_HPP_HEADER_INCLUDED
#define COMPLEX_HPP_HEADER_INCLUDED

#include <complex>
#include <boost/math/constants/constants.hpp>
#include <boost/utility/enable_if.hpp>

namespace boost
{
namespace math
{

/// Forward declarations.
template<typename T> class complex;
template<typename T> class imaginary;

/// Complex number class
template<typename T>
class complex
{
public:

    /// Template type
    typedef T value_type;

    /// Constructs a complex number of the form z = re + i*im
    complex(const T& re = T(), const T& im = T())
        :m_real(re), m_imag(im) {}

    /// Constructs a complex number from a std::complex
    template<typename X>
    explicit complex(const std::complex<X>& z)
        :m_real(z.real()), m_imag(z.imag()) {}

    /// Constructs a number of the form z = 0 + i*im
    template<typename X>
    explicit complex(const imaginary<X>& im)
        :m_real(T(0.)), m_imag(im.imag()) {}

    /// Constructs a number of the form z = re + i*im
    template<typename X>
    explicit complex(const X& re, const imaginary<X>& im)
        :m_real(re), m_imag(im.imag()) {}

    /// Copy constructor
    template<typename X>
    explicit complex(const complex<X>& z)
        :m_real(z.real()), m_imag(z.imag()) {}

    /// Assignment operator
    template<typename X>
    complex<T>& operator=(const complex<X>& z)
    {
        m_real = z.real();
        m_imag = z.imag();
        return *this;
    }

    /// Assignment operator
    complex<T>& operator=(const T& re)
    {
        m_real = re;
        m_imag = T(0.);
        return *this;
    }

    /// Assignment operator
    template<typename X>
    complex<T>& operator=(const imaginary<X>& im)
    {
        m_real = T(0.);
        m_imag = im.imag();
        return *this;
    }

    /// Returns the real part of the number
    T real() const
    {
        return m_real;
    }

    /// Returns the imaginary part of the number
    T imag() const
    {
        return m_imag;
    }

    /// Assigns a new value to the real part
    void real(T val)
    {
        m_real = val;
    }

    /// Assigns a new value to the imaginary part
    void imag(T val)
    {
        m_imag = val;
    }

    /// Addition assignment operator
    complex<T>& operator+=(const T& x)
    {
        m_real += x;
        return *this;
    }

    /// Addition assignment operator
    template<typename X>
    complex<T>& operator+=(const imaginary<X>& x)
    {
        m_imag += x.imag();
        return *this;
    }

    /// Addition assignment operator
    template<typename X>
    complex<T>& operator+=(const complex<X>& z)
    {
        m_real += z.real();
        m_imag += z.imag();
        return *this;
    }

    /// Subtraction assignment operator
    complex<T>& operator-=(const T& x)
    {
        m_real -= x;
        return *this;
    }

    /// Subtraction assignment operator
    template<typename X>
    complex<T>& operator-=(const imaginary<X>& x)
    {
        m_imag -= x.imag();
        return *this;
    }

    /// Subtraction assignment operator
    template<typename X>
    complex<T>& operator-=(const complex<X>& z)
    {
        m_real -= z.real();
        m_imag -= z.imag();
        return *this;
    }

    /// Multiplication assignment operator
    complex<T>& operator*=(const T& x)
    {
        m_real *= x;
        m_imag *= x;
        return *this;
    }

    /// Multiplication assignment operator
    template<typename X>
    complex<T>& operator*=(const imaginary<X>& x)
    {
        const T temp = m_real;
        m_real = -m_imag * x.imag();
        m_imag = temp * x.imag();
        return *this;
    }

    /// Multiplication assignment operator
    template<typename X>
    complex<T>& operator*=(const complex<X>& x)
    {
        const T r = m_real * x.real() - m_imag * x.imag();
        m_imag = m_real * x.imag() + m_imag * x.real();
        m_real = r;
        return *this;
    }

    /// Division assignment operator
    complex<T>& operator/=(const T& x)
    {
        m_real /= x;
        m_imag /= x;
        return *this;
    }

    /// Division assignment operator
    template<typename X>
    complex<T>& operator/=(const imaginary<X>& x)
    {
        const T temp = m_imag;
        m_imag = -m_real / x.imag();
        m_real = temp / x.imag();
        return *this;
    }

    /// Division assignment operator
    template<typename X>
    complex<T>& operator/=(const complex<X>& x)
    {
        const T r =  m_real * x.real() + m_imag * x.imag();
        const T n = norm(x);
        m_imag = (m_imag * x.real() - m_real * x.imag()) / n;
        m_real = r / n;
        return *this;
    }

private:

    /// Implementation - value of the real and imaginary parts of z = a + ib
    T m_real;
    T m_imag;
};

/// Imaginary number class
template<typename T>
class imaginary
{
public:

    /// Template type
    typedef T value_type;

    /// Constructs a number of the form z = 0 + i*im
    explicit imaginary(const T& im = T())
        :m_imag(im) {}

    /// Copy constructor
    template<typename X>
    explicit imaginary(const imaginary<X>& z)
        :m_imag(z.imag()) {}

    /// Assignment operator
    template<typename X>
    imaginary<T>& operator=(const imaginary<X>& z)
    {
        m_imag = z.imag();
        return *this;
    }

    /// Returns the real part of the number (0 by definition)
    T real() const
    {
        return T(0.);
    }

    /// Returns the imaginary part of the number
    T imag() const
    {
        return m_imag;
    }

    /// Assigns a new value to the imaginary part
    void imag(const T& val)
    {
        m_imag = val;
    }

    /// Addition assignment operator
    template<typename X>
    imaginary<T>& operator+=(const imaginary<X>& z)
    {
        m_imag += z.imag();
        return *this;
    }

    /// Subtraction assignment operator
    template<typename X>
    imaginary<T>& operator-=(const imaginary<X>& z)
    {
        m_imag -= z.imag();
        return *this;
    }

    /// Multiplication assignment operator
    imaginary<T>& operator*=(const T& re)
    {
        m_imag *= re;
        return *this;
    }

    /// Division assignment operator
    imaginary<T>& operator/=(const T& re)
    {
        m_imag /= re;
        return *this;
    }

private:

    /// Implementation - value of the imaginary part of z = 0 + ib
    T m_imag;
};


// Addition operators --------------------------------------------------------


/// Returns the value of x plus y
template<typename T>
inline complex<T> operator+(const complex<T>& x, const complex<T>& y)
{
    return complex<T>(x) += y;
}

/// Returns the value of x plus y
template<typename T>
inline complex<T> operator+(const complex<T>& x, const T& y)
{
    return complex<T>(x) += y;
}

/// Returns the value of x plus y
template<typename T>
inline complex<T> operator+(const T& x, const complex<T>& y)
{
    return complex<T>(y) += x;
}

/// Returns the value of x plus built-in arithmetic type val
template <typename Float, typename Arithmetic>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator+(const complex<Float>& x, const Arithmetic& val)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    return result_type(x.real() + val, x.imag());
}

/// Returns the value of built-in arithmetic type val plus x
template <typename Arithmetic, typename Float>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator+(const Arithmetic& val, const complex<Float>& x)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    return result_type(val + x.real(), x.imag());
}

/// Returns the value of x plus y
template<typename T>
inline imaginary<T> operator+(const imaginary<T>& x, const imaginary<T>& y)
{
    return imaginary<T>(x) += y;
}

/// Returns the value of x plus y
template<typename T>
inline complex<T> operator+(const imaginary<T>& x, const T& y)
{
    return complex<T>(y,x);
}

/// Returns the value of x plus y
template<typename T>
inline complex<T> operator+(const T& x, const imaginary<T>& y)
{
    return complex<T>(x,y);
}

/// Returns the value of x plus y
template<typename T>
inline complex<T> operator+(const complex<T>& x, const imaginary<T>& y)
{
    return complex<T>(x) += y;
}

/// Returns the value of x plus y
template<typename T>
inline complex<T> operator+(const imaginary<T>& x, const complex<T>& y)
{
    return complex<T>(y) += x;
}


// Subtraction operators ----------------------------------------------------


/// Returns the value of x minus y
template<typename T>
inline complex<T> operator-(const complex<T>& x, const complex<T>& y)
{
    return complex<T>(x) -= y;
}

/// Returns the value of x minus y
template<typename T>
inline complex<T> operator-(const complex<T>& x, const T& y)
{
    return complex<T>(x) -= y;
}

/// Returns the value of x minus y
template<typename T>
inline complex<T> operator-(const T& x, const complex<T>& y)
{
    return complex<T>(x - y.real(), -y.imag());
}

/// Returns the value of x minus built-in arithmetic type val
template <typename Float, typename Arithmetic>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator-(const complex<Float>& x, const Arithmetic& val)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    return result_type(x.real() - val, x.imag());
}

/// Returns the value of built-in arithmetic type val minus x
template <typename Arithmetic, typename Float>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator-(const Arithmetic& val, const complex<Float>& x)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    return result_type(val - x.real(), -x.imag());
}

/// Returns the value of x minus y
template<typename T>
inline imaginary<T> operator-(const imaginary<T>& x, const imaginary<T>& y)
{
    return imaginary<T>(x) -= y;
}

/// Returns the value of x minus y
template<typename T>
inline complex<T> operator-(const imaginary<T>& x, const T& y)
{
    return complex<T>(-y, x.imag());
}

/// Returns the value of x minus y
template<typename T>
inline complex<T> operator-(const T& x, const imaginary<T>& y)
{
    return complex<T>(x, -y.imag());
}

/// Returns the value of x minus y
template<typename T>
inline complex<T> operator-(const complex<T>& x, const imaginary<T>& y)
{
    return complex<T>(x) -= y;
}

/// Returns the value of x minus y
template<typename T>
inline complex<T> operator-(const imaginary<T>& x, const complex<T>& y)
{
    return complex<T>(-y.real(), x.imag() - y.imag());
}


// Multiplication operators --------------------------------------------------


/// Returns the value of x times y
template<typename T>
inline complex<T> operator*(const complex<T>& x, const complex<T>& y)
{
    return complex<T>(x) *= y;
}

/// Returns the value of x times y
template<typename T>
inline complex<T> operator*(const complex<T>& x, const T& y)
{
    return complex<T>(x) *= y;
}

/// Returns the value of x times y
template<typename T>
inline complex<T> operator*(const T& x, const complex<T>& y)
{
    return complex<T>(y) *= x;
}

/// Returns the value of x times built-in arithmetic type val
template <typename Float, typename Arithmetic>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator*(const complex<Float>& x, const Arithmetic& val)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    return result_type(x.real() * val, x.imag() * val);
}

/// Returns the value of built-in arithmetic type val times x
template <typename Arithmetic, typename Float>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator*(const Arithmetic& val, const complex<Float>& x)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    return result_type(x.real() * val, x.imag() * val);
}

/// Returns the value of x times y
template<typename T>
inline T operator*(const imaginary<T>& x, const imaginary<T>& y)
{
    return -x.imag()*y.imag();
}

/// Returns the value of x times y
template<typename T>
inline imaginary<T> operator*(const imaginary<T>& x, const T& y)
{
    return imaginary<T>(x) *= y;
}

/// Returns the value of x times y
template<typename T>
inline imaginary<T> operator*(const T& x, const imaginary<T>& y)
{
    return imaginary<T>(y) *= x;
}

/// Returns the value of x times y
template<typename T>
inline complex<T> operator*(const complex<T>& x, const imaginary<T>& y)
{
    return complex<T>(x) *= y;
}

/// Returns the value of x times y
template<typename T>
inline complex<T> operator*(const imaginary<T>& x, const complex<T>& y)
{
    return complex<T>(y) *= x;
}


// Division operators --------------------------------------------------------


/// Returns the value of x divided by y
template<typename T>
inline complex<T> operator/(const complex<T>& x, const complex<T>& y)
{
    return complex<T>(x) /= y;
}

/// Returns the value of x divided by y
template<typename T>
inline complex<T> operator/(const complex<T>& x, const T& y)
{
    return complex<T>(x) /= y;
}

/// Returns the value of x divided by y
template<typename T>
inline complex<T> operator/(const T& x, const complex<T>& y)
{
    const T r =  x * y.real();
    const T n = norm(y);
    return complex<T>(r/n , -x*y.imag()/n);
}

/// Returns the value of x divided by y
template<typename T>
inline T operator/(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag()/y.imag();
}

/// Returns the value of x divided by built-in arithmetic type val
template <typename Float, typename Arithmetic>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator/(const complex<Float>& x, const Arithmetic& val)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    return result_type(x.real() / val, x.imag() / val);
}

/// Returns the value of built-in arithmetic type val divided by x
template <typename Arithmetic, typename Float>
inline typename enable_if<is_arithmetic<Arithmetic>, complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> >::type
operator/(const Arithmetic& val, const complex<Float>& x)
{
    typedef complex<typename boost::math::tools::promote_args<Float, Arithmetic>::type> result_type;
    const result_type::value_type n = norm(x);
    return result_type((x.real() * val) / n, - (x.imag() * val) / n);
}

/// Returns the value of x divided by y
template<typename T>
inline imaginary<T> operator/(const imaginary<T>& x, const T& y)
{
    return imaginary<T>(x) /= y;
}

/// Returns the value of x divided by y
template<typename T>
inline imaginary<T> operator/(const T& x, const imaginary<T>& y)
{
    return imaginary<T>(-x/y.imag());
}

/// Returns the value of x divided by y
template<typename T>
inline complex<T> operator/(const complex<T>& x, const imaginary<T>& y)
{
    return complex<T>(x.imag() / y.imag(), - x.real() / y.imag());
}

/// Returns the value of x divided by y
template<typename T>
inline complex<T> operator/(const imaginary<T>& x, const complex<T>& y)
{
    const T n = norm(y);
    return complex<T>(x.imag() * y.imag() / n, x.imag()*y.real() / n);
}


// Arithmetic unary operators ------------------------------------------------


/// Returns x
template<typename T>
inline complex<T> operator+(const complex<T>& x)
{
    return x;
}

/// Returns x
template<typename T>
inline imaginary<T> operator+(const imaginary<T>& x)
{
    return x;
}

/// Returns the opposite of x
template<typename T>
inline complex<T> operator-(const complex<T>& x)
{
    return complex<T>(-x.real(), -x.imag());
}

/// Returns the opposite of x
template<typename T>
inline imaginary<T> operator-(const imaginary<T>& x)
{
    return imaginary<T>(-x.imag());
}


// Comparison operators ------------------------------------------------------

/// Returns true if x equals y
template<typename T>
inline bool operator==(const complex<T>& x, const complex<T>& y)
{
    return x.real() == y.real() && x.imag() == y.imag();
}

/// Returns true if x equals y
template<typename T>
inline bool operator==(const complex<T>& x, const T& y)
{
    return x.real() == y && x.imag() == T(0.);
}

/// Returns true if x equals y
template<typename T>
inline bool operator==(const T& x, const complex<T>& y)
{
    return x == y.real() && y.imag() == T(0.);
}

/// Returns true if x equals y
template<typename T>
inline bool operator==(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag() == y.imag();
}

/// Returns true if x equals y
template<typename T>
inline bool operator==(const imaginary<T>& x, const complex<T>& y)
{
    return x.imag() == y.imag() && y.real() == T(0.);
}

/// Returns true if x equals y
template<typename T>
inline bool operator==(const complex<T>& x, const imaginary<T>& y)
{
    return x.imag() == y.imag() && x.real() == T(0.);
}

/// Returns true if x does not equal y
template<typename T>
inline bool operator!=(const complex<T>& x, const complex<T>& y)
{
    return x.real() != y.real() || x.imag() != y.imag();
}

/// Returns true if x does not equal y
template<typename T>
inline bool operator!=(const complex<T>& x, const T& y)
{
    return x.real() != y || x.imag() != T(0.);
}

/// Returns true if x does not equal y
template<typename T>
inline bool operator!=(const T& x, const complex<T>& y)
{
    return x != y.real() || T(0.) != y.imag();
}

/// Returns true if x does not equal y
template<typename T>
inline bool operator!=(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag() != y.imag();
}

/// Returns true if x does not equal y
template<typename T>
inline bool operator!=(const imaginary<T>& x, const complex<T>& y)
{
    return x.imag() != y.imag() || y.real() != T(0.);
}

/// Returns true if x does not equal y
template<typename T>
inline bool operator!=(const complex<T>& x, const imaginary<T>& y)
{
    return x.imag() != y.imag() || x.real() != T(0.);
}

/// Returns true if x is smaller than y
template<typename T>
inline bool operator<(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag() < y.imag();
}

/// Returns true if x is greater than y
template<typename T>
inline bool operator>(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag() > y.imag();
}

/// Returns true if x is smaller or equal to y
template<typename T>
inline bool operator<=(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag() <= y.imag();
}

/// Returns true if x is greater or equal to y
template<typename T>
inline bool operator>=(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag() >= y.imag();
}


// Stream injection/extraction operators --------------------------------------

/**
 * \brief Stream insertion operator
 *
 * A number z = a + i*b is written as follows: (a, b)
 */
template<typename T, typename charT, class traits>
std::basic_ostream<charT, traits>&
operator<<(std::basic_ostream<charT, traits>& os, const complex<T>& x)
{
    std::basic_ostringstream<charT, traits> s;
    s.flags(os.flags());
    s.imbue(os.getloc());
    s.precision(os.precision());
    s << '(' << x.real() << ',' << x.imag() << ')';
    return os << s.str();
}

/**
 * \brief Stream insertion operator
 *
 * Numbers are written as complex<> numbers are. This means that a number\n
 * a*i will be written as follows: (0., a)
 */
template<typename T, typename charT, class traits>
std::basic_ostream<charT, traits>&
operator<<(std::basic_ostream<charT, traits>& os, const imaginary<T>& x)
{
    std::basic_ostringstream<charT, traits> s;
    s.flags(os.flags());
    s.imbue(os.getloc());
    s.precision(os.precision());
    s << '(' << T(0.) << ',' << x.imag() << ')';
    return os << s.str();
}

/**
 * \brief Stream extraction operator
 *
 * Numbers of the following form will be accepted:\n
 * (a, b)\n
 * (a)\n
 * a\n
 * and will produce a complex number whose value is z = a + i*b.\n
 * Any other input will raise the stream failbit.
 *
 */
template<typename T, typename charT, class traits>
std::basic_istream<charT, traits>&
operator>>(std::basic_istream<charT, traits>& is, complex<T>& x)
{
    T re, im;
    charT temp;
    is >> temp;
    if (temp == '(')
    {
        is >> re >> temp;
        if (temp == ',')
        {
            is >> im >> temp;
            if (temp == ')')
                x = complex<T>(re, im);
            else
                is.setstate(std::ios_base::failbit);
        }
        else if (temp == ')')
            x = re;
        else
            is.setstate(std::ios_base::failbit);
    }
    else
    {
        is.putback(temp);
        is >> re;
        x = re;
    }
    return is;
}

/**
 * \brief Stream extraction operator
 *
 * Numbers are read as complex>> number are.\n
 * This means that numbers of the following form will be accepted:\n
 * (0., a)\n
 * a\n
 * and will produce an imaginary number whose value is a.\n
 * Any other input will raise the stream failbit. \n
 *
 * Note that if the parenthesis notation is used and the first number read is
 * not 0 (e.g. (3., 4.)) the failbit flag of the stream will be raised.
 */
template<typename T, typename charT, class traits>
std::basic_istream<charT, traits>&
operator>>(std::basic_istream<charT, traits>& is, imaginary<T>& x)
{
    T im;
    charT temp;
    is >> temp;
    if (temp == '(')
    {
        is >> im >> temp;
        if (temp == ',' && im == T())
        {
            is >> im >> temp;
            if (temp == ')')
                x = imaginary<T>(im);
            else
                is.setstate(std::ios_base::failbit);
        }
        else
            is.setstate(std::ios_base::failbit);
    }
    else
    {
        is.putback(temp);
        is >> im;
        x = im;
    }
    return is;
}


// Basic complex functions ----------------------------------------------------


/// Returns the real part of a complex number
template<typename T>
inline T real(const complex<T>& x)
{
    return x.real();
}

/// Returns the real part of a pure imaginary number (0 by definition)
template<typename T>
inline T real(const imaginary<T>&)
{
    return T(0.);
}

/// Returns the real part of a real number
template<typename T>
inline T real(const T& x)
{
    return x;
}

/// Returns the imaginary part of a complex number
template<typename T>
inline T imag(const complex<T>& x)
{
    return x.imag();
}


/// Returns the imaginary part of an imaginary number
template<typename T>
inline T imag(const imaginary<T>& x)
{
    return x.imag();
}

/// Returns the imaginary part of a real number (0 by definition)
template<typename T>
inline T imag(const T& x)
{
    return T(0.);
}

/**
 * \brief Returns the magnitude of x
 *
 * The mathematical function sqrt(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline T abs(const complex<T>& x)
{
    return sqrt(x.real() * x.real() + x.imag() * x.imag());
}

/**
 * \brief Returns the magnitude of x
 *
 * The mathematical function abs(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline T abs(const imaginary<T>& x)
{
    return abs(x.imag());
}

/**
 * \brief Returns the argument of x
 *
 * The mathematical functions atan(T) and sqrt(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline T arg(const complex<T>& x)
{
    if(x.real() < T(0.) && x.imag() == T(0.))
        return boost::math::constants::pi<T>();
    return T(2.)*atan(x.imag() / (abs(x) + x.real()));
}

/// Returns the phase angle of x
template<typename T>
inline T arg(const imaginary<T>& x)
{
    return (x.imag() > T(0.)) ?
           boost::math::constants::pi<T>() / 2 :
           ((x.imag() < T(0.)) ?
            T(-boost::math::constants::pi<T>() / 2)
            : T(0.));
}

/// Returns the norm (squared magnitude) of x
template<typename T>
inline T norm(const complex<T>& x)
{
    return x.real() * x.real() + x.imag() * x.imag();
}

/// Returns the norm (squared magnitude) of x
template<typename T>
inline T norm(const imaginary<T>& x)
{
    return x.imag() * x.imag();
}

/// Returns the conjugate of x
template<typename T>
inline complex<T> conj(const complex<T>& x)
{
    return complex<T>(x.real(), -x.imag());
}

/// Returns the conjugate of x
template<typename T>
inline imaginary<T> conj(const imaginary<T>& x)
{
    return imaginary<T>(-x.imag());
}


// Polar representation -------------------------------------------------------


/**
 * \brief Returns the complex number given its polar representation
 * z = r*exp(i*theta)
 *
 * The mathematical functions cos(T) and sin(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> polar(const T& r, const T& theta)
{
    return complex<T>(r*cos(theta), r*sin(theta));
}


// Projection on the Riemann sphere -------------------------------------------


/**
 * \brief Returns the projection of x on the Riemann sphere
 *
 * Specialisations for the native types float, double and long double
 * are provided.
 */
template<typename T>
inline complex<T> proj(const complex<T>& x)
{
    const T ratio = norm(x) + T(1.);
    return complex<T>(T(2.) * x.real() / ratio, T(2.) * x.imag() / ratio);
}


/**
 * \brief Returns the projection of x on the Riemann sphere
 *
 * Specialisations for the native types float, double and long double
 * are provided.
 */
template<typename T>
inline imaginary<T> proj(const imaginary<T>& x)
{
    const T ratio = norm(x) + T(1.);
    return imaginary<T>(T(0.), T(2.) * x.imag() / ratio);
}


// Trigonometric complex functions --------------------------------------------


/**
 * \brief Returns the cosine of x
 *
 * The mathematical functions cos(T), sin(T), cosh(T) and sinh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> cos(const complex<T>& x)
{
    return complex<T>(cos(x.real())*cosh(x.imag()),
                      -sin(x.real())*sinh(x.imag()));
}

/**
 * \brief Returns the cosine of x
 *
 * The mathematical function cosh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline T cos(const imaginary<T>& x)
{
    return cosh(x.imag());
}

/**
 * \brief Returns the sine of x
 *
 * The mathematical functions cos(T), sin(T), cosh(T) and sinh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> sin(const complex<T>& x)
{
    return complex<T>(sin(x.real())*cosh(x.imag()),
                      cos(x.real())*sinh(x.imag()));
}

/**
 * \brief Returns the sine of x
 *
 * The mathematical functions sinh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline imaginary<T> sin(const imaginary<T>& x)
{
    return imaginary<T>(sinh(x.imag()));
}

/**
 * \brief Returns the tangent of x
 *
 * The mathematical functions cos(T), sin(T), cosh(T) and sinh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> tan(const complex<T>& x)
{
    const complex<T> e1 = exp(complex<T>(-x.imag(), x.real()));
    const complex<T> e2 = 1 / e1;
    const complex<T> tan_result = (e1 - e2) / (e1 + e2);
    return complex<T>(tan_result.imag(), -tan_result.real());
}

/**
 * \brief Returns the tangent of x
 *
 * The mathematical functions tanh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline imaginary<T> tan(const imaginary<T>& x)
{
    return imaginary<T>(tanh(x.imag()));
}


// Hyperbolic trigonometric complex functions ---------------------------------


/**
 * \brief Returns the hyperbolic cosine of x
 *
 * The mathematical functions cos(T), sin(T), cosh(T) and sinh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> cosh(const complex<T>& x)
{
    return complex<T>(cosh(x.real())*cos(x.imag()),
                      sinh(x.real())*sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic cosine of x
 *
 * The mathematical functions cos(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline T cosh(const imaginary<T>& x)
{
    return cos(x.imag());
}

/**
 * \brief Returns the hyperbolic sine of x
 *
 * The mathematical functions cos(T), sin(T), cosh(T) and sinh(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> sinh(const complex<T>& x)
{
    return complex<T>(sinh(x.real())*cos(x.imag()),
                      cosh(x.real())*sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic sine of x
 *
 * The mathematical functions sin(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline imaginary<T> sinh(const imaginary<T>& x)
{
    return imaginary<T>(sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic tangent of x
 *
 * The mathematical functions tan(T) must be
 * overloaded for the type T. \n
 * A specialisation for the native types float, double and long double are
 * provided.
 */
template<typename T>
inline complex<T> tanh(const complex<T>& x)
{
    const complex<T> e1 = exp(x);
    const complex<T> e2 = 1 / e1;
    return ( e1 - e2) / (e1 + e2);
}

/**
 * \brief Returns the hyperbolic tangent of x
 *
 * The mathematical functions tan(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline imaginary<T> tanh(const imaginary<T>& x)
{
    return imaginary<T>(tan(x.imag()));
}


// Exponential complex functions ----------------------------------------------


/**
 * \brief Returns the exponential of x
 *
 * The mathematical functions sin(T), cos(T) and exp(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> exp(const complex<T>& x)
{
    return polar<T>(exp(x.real()), x.imag());
}

/**
 * \brief Returns the exponential of x
 *
 * The mathematical functions sin(T) and cos(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> exp(const imaginary<T>& x)
{
    return polar<T>(T(1.), x.imag());
}


// Logarithm complex functions ------------------------------------------------


/**
 * \brief Returns the natural logarithm of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical function log(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> log(const complex<T>& x)
{
    return complex<T>(log(abs(x)), arg(x));
}

/**
 * \brief Returns the natural logarithm of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical function log(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> log(const imaginary<T>& x)
{
    return complex<T>(log(abs(x)), arg(x));
}

/**
 * \brief Returns the base 10 logarithm of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical function log(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> log10(const complex<T>& x)
{
    static const T my_ln_ten = log(T(10));
    return log(x) / my_ln_ten;

    // TBD: Boost.Math.Constants potentially needs ln_ten.
//    return log(x) / boost::math::constants::ln_ten<T>();
}

/**
 * \brief Returns the base 10 logarithm of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical function log(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> log10(const imaginary<T>& x)
{
    static const T my_ln_ten = log(T(10));
    return log(x) / my_ln_ten;

    // TBD: Boost.Math.Constants potentially needs ln_ten.
//    return log(x) / boost::math::constants::ln_ten<T>();
}

/**
 * \brief Returns the base 2 logarithm of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical function log(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> log2(const complex<T>& x)
{
    return log(x) / boost::math::constants::ln_two<T>();
}

/**
 * \brief Returns the base 2 logarithm of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical function log(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> log2(const imaginary<T>& x)
{
    return log(x) / boost::math::constants::ln_two<T>();
}


// Square root complex function -----------------------------------------------


/**
 * \brief Returns the square root of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical functions abs(T) and sqrt(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> sqrt(const complex<T>& x)
{
    const T t = sqrt(T(2.) * (abs(x) + abs(x.real())));
    const T u = t * T(0.5);
    return x.real() > T(0.) ?
           complex<T>(u, x.imag() / t) :
           complex<T>(abs(x.imag()) / t, x.imag() < T() ? -u : u);
}


/**
 * \brief Returns the square root of x
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical functions abs(T) and sqrt(T) must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> sqrt(const imaginary<T>& x)
{
    const T val = sqrt(abs(x.imag()) * T(0.5));
    return complex<T>(val, x.imag() < T() ? -val : val);
}


// Complex power functions ----------------------------------------------------



/**
 * \brief Returns the complex power base of x raised to the n-th power.
 *
 * The branch cuts are on the negative real axis. \n
 */
template<typename T>
complex<T> pow(complex<T> x, unsigned int n)
{
    complex<T> y = n % 2 ? x : complex<T>(1.);
    while (n >>= 1)
    {
        x *= x;
        if (n % 2)
            y *= x;
    }
    return y;
}

/**
 * \brief Returns the complex power base of x raised to the n-th power.
 *
 * The branch cuts are on the negative real axis. \n
 */
template<typename T>
inline complex<T> pow(const complex<T>& x, int n)
{
    return n < 0
           ? complex<T>(1.) / pow(x, static_cast<unsigned int>(-n))
           : pow(x, static_cast<unsigned int>(n));
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 */
template<typename T>
inline complex<T> pow(const complex<T>& x, const complex<T>& y)
{
    return x == T(0.) ? complex<T>(T(0.),T(0.)) : exp(y * log(x));
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical functions exp(T), sin(T) and cos(T)  must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> pow(const complex<T>& x, const T& y)
{
    if(x == T(0.))
        return T(0.);

    const complex<T> t = log(x);
    return polar(exp(y * t.real()), y * t.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical functions pow(T,T), log(T), sin(T) and cos(T)  must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> pow(const T& x, const complex<T>& y)
{
    return  x > T(0.) ?
            polar(pow(x, y.real()), y.imag() * log(x)) :
            pow(complex<T>(x), y);
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 */
template<typename T>
inline complex<T> pow(const imaginary<T>& x, const imaginary<T>& y)
{
    return x.imag() == T(0.) ? complex<T>(T(0.), T(0.)) : exp(y * log(x));
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical functions exp(T), sin(T) and cos(T)  must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> pow(const imaginary<T>& x, const T& y)
{
    const complex<T> t = log(x);
    return polar(exp(y*t.real()), y*t.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 * The mathematical functions log(T), sin(T) and cos(T)  must be
 * overloaded for the type T. \n
 * Specialisations for the native types float, double and long double
 * using the SL mathematical functions are provided.
 */
template<typename T>
inline complex<T> pow(const T& x, const imaginary<T>& y)
{
    return  x > T(0.) ?
            polar(T(1.), y.imag() * log(x)) :
            pow(complex<T>(x), y);
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 */
template<typename T>
inline complex<T> pow(const imaginary<T>& x, const complex<T>& y)
{
    return x.imag() == T(0.) ? complex<T>(T(0.), T(0.)) : exp(y * log(x));
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 *
 * The branch cuts are on the negative real axis. \n
 */
template<typename T>
inline complex<T> pow(const complex<T>& x, const imaginary<T>& y)
{
    return x == T(0.) ? complex<T>(T(0.), T(0.)) : exp(y * log(x));
}





// SPECIALISATIONS ############################################################



// Basic complex functions specialisation  ------------------------------------


/// Returns the magnitude of x (specialisation for float type)
template<>
inline float abs(const complex<float>& x)
{
    return std::sqrt(x.real() * x.real() + x.imag() * x.imag());
}

/// Returns the magnitude of x (specialisation for double type)
template<>
inline double abs(const complex<double>& x)
{
    return std::sqrt(x.real() * x.real() + x.imag() * x.imag());
}

/// Returns the magnitude of x (specialisation for long double type)
template<>
inline long double abs(const complex<long double>& x)
{
    return std::sqrt(x.real() * x.real() + x.imag() * x.imag());
}

/// Returns the magnitude of x (specialisation for float type)
template<>
inline float abs(const imaginary<float>& x)
{
    return std::abs(x.imag());
}

/// Returns the magnitude of x (specialisation for double type)
template<>
inline double abs(const imaginary<double>& x)
{
    return std::abs(x.imag());
}

/// Returns the magnitude of x (specialisation for long double type)
template<>
inline long double abs(const imaginary<long double>& x)
{
    return std::abs(x.imag());
}

/// Returns the argument of x (specialisation for float type)
template<>
inline float arg(const complex<float>& x)
{
    return std::atan2(x.imag(), x.real());
}

/// Returns the argument of x (specialisation for double type)
template<>
inline double arg(const complex<double>& x)
{
    return std::atan2(x.imag(), x.real());
}

/// Returns the argument of x (specialisation for long double type)
template<>
inline long double arg(const complex<long double>& x)
{
    return std::atan2(x.imag(), x.real());
}


// Polar representation specialisation ----------------------------------------


/**
 * \brief Returns the complex number given its polar representation
 * z=r*exp(i*theta) (specialisation for float type)
 */
template<>
inline complex<float> polar(const float& r, const float& theta)
{
    return complex<float>(r*std::cos(theta), r*std::sin(theta));
}

/**
 * \brief Returns the complex number given its polar representation
 * z=r*exp(i*theta) (specialisation for double type)
 */
template<>
inline complex<double> polar(const double& r, const double& theta)
{
    return complex<double>(r*std::cos(theta), r*std::sin(theta));
}

/**
 * \brief Returns the complex number given its polar representation
 * z=r*exp(i*theta) (specialisation for long double type)
 */
template<>
inline complex<long double> polar(const long double& r,
                                  const long double& theta)
{
    return complex<long double>(r*std::cos(theta), r*std::sin(theta));
}


// Projection on the Riemann sphere specialisation ----------------------------


/**
 * \brief Returns the projection of x on the Riemann sphere
 * (specialisation for float type)
 */
template<>
inline complex<float> proj(const complex<float>& x)
{
    const float ratio = norm(x) + 1.f;
    return complex<float>(2.f * x.real() / ratio, 2.f * x.imag() / ratio);
}

/**
 * \brief Returns the projection of x on the Riemann sphere
 * (specialisation for float type)
 */
template<>
inline imaginary<float> proj(const imaginary<float>& x)
{
    const float ratio = norm(x) + 1.f;
    return imaginary<float>(2.f * x.imag() / ratio);
}

/**
 * \brief Returns the projection of x on the Riemann sphere
 * (specialisation for double type)
 */
template<>
inline complex<double> proj(const complex<double>& x)
{
    const double ratio = norm(x) + 1.;
    return complex<double>(2. * x.real() / ratio, 2. * x.imag() / ratio);
}

/**
 * \brief Returns the projection of x on the Riemann sphere
 * (specialisation for double type)
 */
template<>
inline imaginary<double> proj(const imaginary<double>& x)
{
    const double ratio = norm(x) + 1.;
    return imaginary<double>(2. * x.imag() / ratio);
}

/**
 * \brief Returns the projection of x on the Riemann sphere
 * (specialisation for long double type)
 */
template<>
inline complex<long double> proj(const complex<long double>& x)
{
    const long double ratio = norm(x) + 1.l;
    return complex<long double>(2.l * x.real() / ratio, 2.l * x.imag() / ratio);
}

/**
 * \brief Returns the projection of x on the Riemann sphere
 * (specialisation for long double type)
 */
template<>
inline imaginary<long double> proj(const imaginary<long double>& x)
{
    const long double ratio = norm(x) + 1.l;
    return imaginary<long double>(2.l * x.imag() / ratio);
}


// Trigonometric complex function specialisation ------------------------------


/**
 * \brief Returns the cosine of x
 * (specialisation for float type)
 */
template<>
inline complex<float> cos(const complex<float>& x)
{
    const std::complex<float> temp(x.real(), x.imag());
    const std::complex<float> ret = cos(temp);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the cosine of x
 * (specialisation for double type)
 */
template<>
inline complex<double> cos(const complex<double>& x)
{
    const std::complex<double> temp(x.real(), x.imag());
    const std::complex<double> ret = cos(temp);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the cosine of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> cos(const complex<long double>& x)
{
    const std::complex<long double> temp(x.real(), x.imag());
    const std::complex<long double> ret = cos(temp);
    return complex<long double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the cosine of x
 * (specialisation for float type)
 */
template<>
inline float cos(const imaginary<float>& x)
{
    return std::cosh(x.imag());
}

/**
 * \brief Returns the cosine of x
 * (specialisation for double type)
 */
template<>
inline double cos(const imaginary<double>& x)
{
    return std::cosh(x.imag());
}

/**
 * \brief Returns the cosine of x
 * (specialisation for long double type)
 */
template<>
inline long double cos(const imaginary<long double>& x)
{
    return std::cosh(x.imag());
}

/**
 * \brief Returns the sine of x
 * (specialisation for float type)
 */
template<>
inline complex<float> sin(const complex<float>& x)
{
    const std::complex<float> temp(x.real(), x.imag());
    const std::complex<float> ret = sin(temp);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the sine of x
 * (specialisation for double type)
 */
template<>
inline complex<double> sin(const complex<double>& x)
{
    const std::complex<double> temp(x.real(), x.imag());
    const std::complex<double> ret = sin(temp);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the sine of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> sin(const complex<long double>& x)
{
    const std::complex<long double> temp(x.real(), x.imag());
    const std::complex<long double> ret = sin(temp);
    return complex<long double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the sine of x
 * (specialisation for float type)
 */
template<>
inline imaginary<float> sin(const imaginary<float>& x)
{
    return imaginary<float>(std::sinh(x.imag()));
}

/**
 * \brief Returns the sine of x
 * (specialisation for double type)
 */
template<>
inline imaginary<double> sin(const imaginary<double>& x)
{
    return imaginary<double>(std::sinh(x.imag()));
}

/**
 * \brief Returns the sine of x
 * (specialisation for long double type)
 */
template<>
inline imaginary<long double> sin(const imaginary<long double>& x)
{
    return imaginary<long double>(std::sinh(x.imag()));
}

/**
 * \brief Returns the tangent of x
 * (specialisation for float type)
 */
template<>
inline imaginary<float> tan(const imaginary<float>& x)
{
    return imaginary<float>(std::tanh(x.imag()));
}

/**
 * \brief Returns the tangent of x
 * (specialisation for double type)
 */
template<>
inline imaginary<double> tan(const imaginary<double>& x)
{
    return imaginary<double>(std::tanh(x.imag()));
}

/**
 * \brief Returns the tangent of x
 * (specialisation for long double type)
 */
template<>
inline imaginary<long double> tan(const imaginary<long double>& x)
{
    return imaginary<long double>(std::tanh(x.imag()));
}

/**
 * \brief Returns the tangent of x
 * (specialisation for float type)
 */
template<>
inline complex<float> tan(const complex<float>& x)
{
    const std::complex<float> temp(x.real(), x.imag());
    const std::complex<float> ret = tan(temp);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the tangent of x
 * (specialisation for double type)
 */
template<>
inline complex<double> tan(const complex<double>& x)
{
    const std::complex<double> temp(x.real(), x.imag());
    const std::complex<double> ret = tan(temp);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the tangent of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> tan(const complex<long double>& x)
{
    const std::complex<long double> temp(x.real(), x.imag());
    const std::complex<long double> ret = tan(temp);
    return complex<long double>(ret.real(), ret.imag());
}



// Hyperbolic trigonometric complex function specialisation -------------------


/**
 * \brief Returns the hyperbolic cosine of x
 * (specialisation for float type)
 */
template<>
inline complex<float> cosh(const complex<float>& x)
{
    return complex<float>(std::cosh(x.real())*std::cos(x.imag()),
                        std::sinh(x.real())*std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic cosine of x
 * (specialisation for double type)
 */
template<>
inline complex<double> cosh(const complex<double>& x)
{
    return complex<double>(std::cosh(x.real())*std::cos(x.imag()),
                        std::sinh(x.real())*std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic cosine of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> cosh(const complex<long double>& x)
{
    return complex<long double>(std::cosh(x.real())*std::cos(x.imag()),
                        std::sinh(x.real())*std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic cosine of x
 * (specialisation for float type)
 */
template<>
inline float cosh(const imaginary<float>& x)
{
    return std::cos(x.imag());
}

/**
 * \brief Returns the hyperbolic cosine of x
 * (specialisation for double type)
 */
template<>
inline double cosh(const imaginary<double>& x)
{
    return std::cos(x.imag());
}

/**
 * \brief Returns the hyperbolic cosine of x
 * (specialisation for long double type)
 */
template<>
inline long double cosh(const imaginary<long double>& x)
{
    return std::cos(x.imag());
}

/**
 * \brief Returns the hyperbolic sine of x
 * (specialisation for float type)
 */
template<>
inline complex<float> sinh(const complex<float>& x)
{
    return complex<float>(std::sinh(x.real())*std::cos(x.imag()),
                      std::cosh(x.real())*std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic sine of x
 * (specialisation for double type)
 */
template<>
inline complex<double> sinh(const complex<double>& x)
{
    return complex<double>(std::sinh(x.real())*std::cos(x.imag()),
                      std::cosh(x.real())*std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic sine of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> sinh(const complex<long double>& x)
{
    return complex<long double>(std::sinh(x.real())*std::cos(x.imag()),
                      std::cosh(x.real())*std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic sine of x
 * (specialisation for float type)
 */
template<>
inline imaginary<float> sinh(const imaginary<float>& x)
{
    return imaginary<float>(std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic sine of x
 * (specialisation for double type)
 */
template<>
inline imaginary<double> sinh(const imaginary<double>& x)
{
    return imaginary<double>(std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic sine of x
 * (specialisation for long double type)
 */
template<>
inline imaginary<long double> sinh(const imaginary<long double>& x)
{
    return imaginary<long double>(std::sin(x.imag()));
}

/**
 * \brief Returns the hyperbolic tangent of x
 * (specialisation for float type)
 */
template<>
inline imaginary<float> tanh(const imaginary<float>& x)
{
    return imaginary<float>(std::tan(x.imag()));
}

/**
 * \brief Returns the hyperbolic tangent of x
 * (specialisation for double type)
 */
template<>
inline imaginary<double> tanh(const imaginary<double>& x)
{
    return imaginary<double>(std::tan(x.imag()));
}

/**
 * \brief Returns the hyperbolic tangent of x
 * (specialisation for long double type)
 */
template<>
inline imaginary<long double> tanh(const imaginary<long double>& x)
{
    return imaginary<long double>(std::tan(x.imag()));
}

/**
 * \brief Returns the hyperbolic tangent of x
 * (specialisation for float type)
 */
template<>
inline complex<float> tanh(const complex<float>& x)
{
    const std::complex<float> temp(x.real(), x.imag());
    const std::complex<float> ret = tanh(temp);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the hyperbolic tangent of x
 * (specialisation for double type)
 */
template<>
inline complex<double> tanh(const complex<double>& x)
{
    const std::complex<double> temp(x.real(), x.imag());
    const std::complex<double> ret = tanh(temp);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the hyperbolic tangent of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> tanh(const complex<long double>& x)
{
    const std::complex<long double> temp(x.real(), x.imag());
    const std::complex<long double> ret = tanh(temp);
    return complex<long double>(ret.real(), ret.imag());
}



// Exponential complex function specialisation --------------------------------


/**
 * \brief Returns the exponential of x
 * (specialisation for float type)
 */
template<>
inline complex<float> exp(const complex<float>& x)
{
    return polar(std::exp(x.real()), x.imag());
}

/**
 * \brief Returns the exponential of x
 * (specialisation for double type)
 */
template<>
inline complex<double> exp(const complex<double>& x)
{
    return polar(std::exp(x.real()), x.imag());
}

/**
 * \brief Returns the exponential of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> exp(const complex<long double>& x)
{
    return polar(std::exp(x.real()), x.imag());
}

/**
 * \brief Returns the exponential of x
 * (specialisation for float type)
 */
template<>
inline complex<float> exp(const imaginary<float>& x)
{
    return polar(1.f, x.imag());
}

/**
 * \brief Returns the exponential of x
 * (specialisation for double type)
 */
template<>
inline complex<double> exp(const imaginary<double>& x)
{
    return polar(1., x.imag());
}

/**
 * \brief Returns the exponential of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> exp(const imaginary<long double>& x)
{
    return polar(1.l, x.imag());
}


// Logarithm complex function specialisation ----------------------------------


/**
 * \brief Returns the natural logarithm of x
 * (specialisation for float type)
 */
template<>
inline complex<float> log(const complex<float>& x)
{
    const std::complex<float> temp(x.real(), x.imag());
    const std::complex<float> ret = std::log(temp);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the natural logarithm of x
 * (specialisation for double type)
 */
template<>
inline complex<double> log(const complex<double>& x)
{
    const std::complex<double> temp(x.real(), x.imag());
    const std::complex<double> ret = std::log(temp);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the natural logarithm of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> log(const complex<long double>& x)
{
    const std::complex<long double> temp(x.real(), x.imag());
    const std::complex<long double> ret = std::log(temp);
    return complex<long double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the natural logarithm of x
 * (specialisation for float type)
 */
template<>
inline complex<float> log(const imaginary<float>& x)
{
    return complex<float>(std::log(abs(x)), arg(x));
}

/**
 * \brief Returns the natural logarithm of x
 * (specialisation for double type)
 */
template<>
inline complex<double> log(const imaginary<double>& x)
{
    return complex<double>(std::log(abs(x)), arg(x));
}

/**
 * \brief Returns the natural logarithm of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> log(const imaginary<long double>& x)
{
    return complex<long double>(std::log(abs(x)), arg(x));
}


// Complex square root function specialisation --------------------------------


/**
 * \brief Returns the square root of x
 * (specialisation for float type)
 */
template<>
inline complex<float> sqrt(const imaginary<float>& x)
{
    const float val = std::sqrt(std::abs(x.imag()) * 0.5f);
    return complex<float>(val, x.imag() < 0.f ? -val : val);
}

/**
 * \brief Returns the square root of x
 * (specialisation for double type)
 */
template<>
inline complex<double> sqrt(const imaginary<double>& x)
{
    const double val = std::sqrt(std::abs(x.imag()) * 0.5);
    return complex<double>(val, x.imag() < 0. ? -val : val);
}

/**
 * \brief Returns the square root of x
 * (specialisation for double type)
 */
template<>
inline complex<long double> sqrt(const imaginary<long double>& x)
{
    const long double val = std::sqrt(std::abs(x.imag()) * 0.5l);
    return complex<long double>(val, x.imag() < 0.l ? -val : val);
}

/**
 * \brief Returns the square root of x
 * (specialisation for float type)
 */
template<>
inline complex<float> sqrt(const complex<float>& x)
{
    const std::complex<float> temp(x.real(), x.imag());
    const std::complex<float> ret = std::sqrt(temp);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the square root of x
 * (specialisation for double type)
 */
template<>
inline complex<double> sqrt(const complex<double>& x)
{
    const std::complex<double> temp(x.real(), x.imag());
    const std::complex<double> ret = std::sqrt(temp);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the square root of x
 * (specialisation for long double type)
 */
template<>
inline complex<long double> sqrt(const complex<long double>& x)
{
    const std::complex<long double> temp(x.real(), x.imag());
    const std::complex<long double> ret = std::sqrt(temp);
    return complex<long double>(ret.real(), ret.imag());
}


// Complex power function specialisation --------------------------------------

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for float type)
 */
template<>
inline complex<float> pow(const complex<float>& x,
                          const complex<float>& y)
{
    const std::complex<float> temp1(x.real(), x.imag());
    const std::complex<float> temp2(y.real(), y.imag());
    const std::complex<float> ret = std::pow(temp1, temp2);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for double type)
 */
template<>
inline complex<double> pow(const complex<double>& x,
                          const complex<double>& y)
{
    const std::complex<double> temp1(x.real(), x.imag());
    const std::complex<double> temp2(y.real(), y.imag());
    const std::complex<double> ret = std::pow(temp1, temp2);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for long double type)
 */
template<>
inline complex<long double> pow(const complex<long double>& x,
                          const complex<long double>& y)
{
    const std::complex<long double> temp1(x.real(), x.imag());
    const std::complex<long double> temp2(y.real(), y.imag());
    const std::complex<long double> ret = std::pow(temp1, temp2);
    return complex<long double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for float type)
 */
template<>
inline complex<float> pow(const complex<float>& x, const float& y)
{
    const std::complex<float> temp(x.real(), x.imag());
    const std::complex<float> ret = std::pow(temp, y);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for double type)
 */
template<>
inline complex<double> pow(const complex<double>& x, const double& y)
{
    const std::complex<double> temp(x.real(), x.imag());
    const std::complex<double> ret = std::pow(temp, y);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for long double type)
 */
template<>
inline complex<long double> pow(const complex<long double>& x,
                                const long double& y)
{
    const std::complex<long double> temp(x.real(), x.imag());
    const std::complex<long double> ret = std::pow(temp, y);
    return complex<long double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for float type)
 */
template<>
inline complex<float> pow(const float& x, const complex<float>& y)
{
    const std::complex<float> temp(y.real(), y.imag());
    const std::complex<float> ret = std::pow(x, temp);
    return complex<float>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for double type)
 */
template<>
inline complex<double> pow(const double& x, const complex<double>& y)
{
    const std::complex<double> temp(y.real(), y.imag());
    const std::complex<double> ret = std::pow(x, temp);
    return complex<double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for long double type)
 */
template<>
inline complex<long double> pow(const long double& x,
                                const complex<long double>& y)
{
    const std::complex<long double> temp(y.real(), y.imag());
    const std::complex<long double> ret = std::pow(x, temp);
    return complex<long double>(ret.real(), ret.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for float type)
 */
template<>
inline complex<float> pow(const imaginary<float>& x, const float& y)
{
    const complex<float> t = log(x);
    return polar(std::exp(y*t.real()), y*t.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for double type)
 */
template<>
inline complex<double> pow(const imaginary<double>& x, const double& y)
{
    const complex<double> t = log(x);
    return polar(std::exp(y*t.real()), y*t.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for long double type)
 */
template<>
inline complex<long double> pow(const imaginary<long double>& x,
                                 const long double& y)
{
    const complex<long double> t = log(x);
    return polar(std::exp(y*t.real()), y*t.imag());
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for float type)
 */
template<>
inline complex<float> pow(const float& x, const imaginary<float>& y)
{
    return  x > 0.f ?
            polar(1.f, y.imag() * std::log(x)) :
            pow(complex<float>(x), y);
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for double type)
 */
template<>
inline complex<double> pow(const double& x, const imaginary<double>& y)
{
    return  x > 0. ?
            polar(1., y.imag() * std::log(x)) :
            pow(complex<double>(x), y);
}

/**
 * \brief Returns the complex power base of x raised to the y-th power.
 * (specialisation for long double type)
 */
template<>
inline complex<long double> pow(const long double& x,
                                const imaginary<long double>& y)
{
    return  x > 0.l ?
            polar(1.l, y.imag() * std::log(x)) :
            pow(complex<long double>(x), y);
}

} //namespace math
} //namespace boost

#endif // COMPLEX_HPP_HEADER_INCLUDED

